import fabric.api as fab

import datetime
import os

from .config import conf

_HIDE = ['output', 'running']


def name(target):
    '''
    Generate a release name with the ISO8601 date and the short rev tag
    from the ref in our repository
    '''
    with fab.hide(*_HIDE):
        now = datetime.datetime.utcnow()
        release_ref = fab.local('git ls-remote {repo_url} {git_ref} | '
                                'head -n 1 | '
                                "awk '{{print substr($1,0,7)}}'".format(
                                    repo_url=conf('repo_url'),
                                    git_ref=conf('git_ref'),
                                ),
                                capture=True)
        release_name = "-".join((
            now.strftime("%Y%m%dT%H%M%S%z"),
            release_ref
        ))

    # set some extra variables
    fab.env.config['variables']['release'] = release_name
    fab.env.config['variables']['release_dir'] = os.path.join(
        conf('releases_dir'),
        release_name
    )

    return release_name


def create(release_name):
    '''
    Create a release on each host with the name specified
    '''
    git_ref = conf('git_ref')
    repo_url = conf('repo_url')
    source_dir = conf('source_dir')
    releases_dir = conf('releases_dir')

    fab.puts("Getting latest commits from our repository...")
    with fab.hide(*_HIDE):
        fab.run("test -d {source_dir} || "
                "git clone --recursive {repo_url} {source_dir}".format(
                    repo_url=repo_url,
                    source_dir=source_dir
                ), pty=False)

    with fab.hide(*_HIDE):
        with fab.cd(source_dir):
            host_repo_url = fab.run("|".join([
                "git remote -v",
                "grep ^origin",
                "head -n 1",
                "awk '{print $2}'"
            ]))

            if host_repo_url != repo_url:
                return fab.abort(
                    "The repository URL doesn't match the URL in our config. "
                    "Cowardly refusing to continue..."
                )

            fab.run("git checkout {git_ref}".format(git_ref=git_ref))
            fab.run("git pull origin")

            # use git to archive it
            fab.puts("Archiving release...")
            fab.run("test -d {releases_dir} || mkdir -p {releases_dir}".format(
                releases_dir=releases_dir
            ))
            fab.run("|".join([
                "git archive --format=tar --prefix={release_name}/ {git_ref}",
                "(cd {releases_dir}; tar xf -)".format(
                    release_name=release_name,
                    git_ref=git_ref,
                    releases_dir=releases_dir)
            ]))

            # now find any submodules
            fab.puts("Looking for submodules...")
            with fab.settings(fab.hide('warnings'), warn_only=True):
                # we hide warnings and don't fail here as xargs returns a
                # non-zero exit status if there is no input passed to it
                git_submodules = fab.run("|".join([
                    "find . -mindepth 2 -name .git -print",
                    "xargs grep -l '^gitdir:'"
                ]))

            for submodule in git_submodules.splitlines():
                submodule = submodule.lstrip("./").rstrip("/.git")

                submodule_ref = fab.run("|".join([
                    "git submodule status %s",
                    "awk '{print $1}'" % submodule
                ]))
                prefix = "{release_name}/{submodule}/".format(
                    submodule=submodule,
                    release_name=release_name,
                )

                # archive it
                fab.puts(" -> Archiving %s..." % submodule)
                fab.run(
                    "("
                    "cd {submodule}; "
                    "git archive --format=tar --prefix={prefix} {submodule_ref}"  # noqa
                    ") | "
                    "(cd {releases_dir}; tar xf -)".format(
                        submodule_ref=submodule_ref,
                        prefix=prefix,
                        releases_dir=releases_dir
                    )
                )


def make_current(release_name):
    '''
    Makes the specified release the current one

    :param release_name: The release name (returned from prepare_release)
    :return:
    '''
    with fab.hide(*_HIDE):
        releases_dir = conf('releases_dir')
        current_symlink = conf('current_symlink')
        fab.run("rm -f {current_symlink} && "
                "ln -s {releases_dir}/{release_name} {current_symlink}".format(
                    current_symlink=current_symlink,
                    releases_dir=releases_dir,
                    release_name=release_name
                ))


def cleanup():
    '''
    Cleans up the release folder

    :return: None
    '''
    keep_releases = conf('keep_releases')
    releases_dir = conf('releases_dir')

    fab.puts("Cleaning up releases folder. Keeping %d latest releases" % (
        keep_releases,
    ))
    with fab.hide(*_HIDE):
        # build the command line to find the releases to cleanup
        commands = [
            "find %s/* -maxdepth 0 -printf '%%T@ %%p\\n'" % releases_dir,
            "sort -k 1 -n",
            "awk '{print $2}'",
            "head -n -%d" % keep_releases,
        ]
        releases = fab.run("|".join(commands)).split()

        for fullpath in releases:
            release = os.path.basename(fullpath)
            fab.puts(" * Removing %s" % release)
            fab.run("rm -fr %s" % fullpath)


def list():
    '''
    Lists all of the releases on the host

    :return: list of releases
    '''
    releases_dir = conf('releases_dir')
    ret = fab.run("/bin/ls {releases_dir}".format(
        releases_dir=releases_dir
    ))
    return ret.split()



import copy
import os
import yaml

import fabric.api as fab

from collections import OrderedDict


class GarmentConfigError(Exception):
    pass


def variable_template(value, variables=None):
    '''
    Apply the config variables to the supplied value
    '''
    if value:
        if not isinstance(value, basestring):
            return value

        if variables is None:
            variables = fab.env.config['variables']

        try:
            return value.format(**variables)
        except KeyError as exc:
            fab.abort('Undefined variable requested: {exc}\n'
                      'Original value we tried to parse: {value}'.format(
                          exc=exc,
                          value=value
                      ))


def conf(name):
    '''
    Shortcut function for getting config items
    '''
    return variable_template(fab.env.config[name])


def load(target, config_file):
    '''
    Load into the environment the target config from the specified file

    :param target: The name of the target to load in the config
    :param config_file: The config file to load, relative to the fabfile
    :return: None
    '''
    if 'config_loaded' in fab.env and \
            fab.env.config_loaded == (target, config_file):
        # we've already loaded our config for this target
        return

    if not config_file.startswith('/'):
        fab_path = os.path.dirname(fab.env.real_fabfile)
        config_file = os.path.join(fab_path, config_file)

    if not os.path.isfile(config_file):
        return fab.abort('No config file found. Looked for: %s' % config_file)

    with open(config_file, 'r') as f:
        try:
            config = yaml.load(f.read())
        except yaml.YAMLError as e:
            line = column = -1
            if hasattr(e, 'problem_mark'):
                line = e.problem_mark.line + 1
                column = e.problem_mark.column + 1
            return fab.abort('Error in %s YAML syntax. Line %d, column %d' % (
                config_file, line, column)
            )

    # make sure our deployment target exists
    if target not in config:
        return fab.abort('The target "%s" is not defined in the config.' % (
            target
        ))

    # are we extending another item in the config
    if 'extends' in config[target]:
        extends = config[target]['extends']

        if extends not in config:
            return fab.abort('The target "%s" requested to extend "%s", '
                             'but "%s" is not defined in the config' % (
                                 target,
                                 extends,
                                 extends
                             ))

        # copy the item we're extending
        extended_config = copy.deepcopy(config[extends])

        # drop hosts, it is never extended
        try:
            del extended_config['hosts']
        except KeyError:
            pass

        # merge our values
        for key in config[target]:
            if key == 'extends':
                # don't copy the extends value
                continue
            elif key == 'variables':
                # if our target sets a variable that's also set in the config
                # its extending then the variable must be set in the same order
                # so that resolution will work properly

                # go through list on target and build a list & dict
                var_names = []
                var_values = {}
                for kvpair in config[target]['variables']:
                    for key, val in kvpair.iteritems():
                        var_names.append(key)
                        var_values[key] = val

                # now go through the extended config
                # when we find a key that exists in var_names we use the value
                # in var_values instead
                if 'variables' not in extended_config:
                    extended_config['variables'] = []

                new_variables = []
                for kvpair in extended_config['variables']:
                    for key, val in kvpair.iteritems():
                        if key in var_names:
                            new_variables.append({key: var_values[key]})

                            # delete it from var_names
                            var_names.remove(key)
                        else:
                            new_variables.append({key: val})

                # finally, anything left in var_names just gets appended
                for key in var_names:
                    new_variables.append({key: var_values[key]})

                extended_config['variables'] = new_variables

            elif key == 'stages':
                # stages are similar to variables, but we need to handle our
                # four stages specifically
                new_stages = {}
                for stage in ['before', 'after', 'cleanup', 'rollback']:

                    if stage not in config[target]['stages']:
                        # if the stage does not exist in our target but does
                        # exist in our extended config then just copy the
                        # whole stage from the extended one
                        if stage in extended_config['stages']:
                            new_stages[stage] = copy.deepcopy(
                                extended_config['stages'][stage]
                            )
                            continue
                    else:
                        # the stage does exist in our target config
                        # if it doesn't exist in the extended config then just
                        # copy the whole stage from the target
                        if stage not in extended_config['stages']:
                            new_stages[stage] = copy.deepcopy(
                                config[target]['stages'][stage]
                            )
                            continue

                        # the stage exists in both configs at this point
                        # merge them
                        stage_names = []
                        stage_data = {}
                        for step in config[target]['stages'][stage]:
                            step_id = None
                            for key, val in step.iteritems():
                                if key == 'id':
                                    step_id = val

                            stage_names.append(step_id)
                            stage_data[step_id] = step

                        new_stage = []

                        # check to see if we're over writing any existing steps
                        for step in extended_config['stages'][stage]:
                            step_id = None
                            for key, val in step.iteritems():
                                if key == 'id':
                                    step_id = val
                                    break

                            if step_id in stage_names:
                                new_stage.append(stage_data[step_id])
                                stage_names.remove(step_id)
                            else:
                                new_stage.append(step)

                        # anything left gets appended
                        for step_id in stage_names:
                            new_stage.append(stage_data[step_id])

                        # set it
                        new_stages[stage] = new_stage

                # set it
                extended_config['stages'] = new_stages

            else:
                extended_config[key] = copy.deepcopy(config[target][key])

        # reset config
        config[target] = extended_config

    # reset our config to the target as the root
    fab.env.config = config[target]
    fab.env.config_loaded = (target, config_file)

    # enable agent forwarding by default
    if 'forward_agent' in config and config['forward_agent'] is False:
        fab.env.forward_agent = False
    else:
        fab.env.forward_agent = True

    # setup our host roles
    if 'hosts' not in fab.env.config:
        return fab.abort('The target "%s" does not define any hosts.' % target)

    roles = {
        'all': []
    }
    for hostname, hostconfig in fab.env.config['hosts'].iteritems():
        roles['all'].append(hostname)

        for role in hostconfig['roles']:
            if role not in roles:
                roles[role] = [hostname]
            else:
                roles[role].append(hostname)

    # set the roles with fabric
    fab.env.roledefs.update(roles)

    variables = OrderedDict([])

    # check required config items
    for name in ('repo_url', 'git_ref', 'deploy_dir'):
        if name not in fab.env.config:
            raise GarmentConfigError('You must supply a "%s" value '
                                     'in the target "%s"' % (
                                         name,
                                         target
                                     ))

        # push it into the variables
        variables[name] = fab.env.config[name]

    # set our defaults
    def default(name, default):
        if name not in fab.env.config:
            fab.env.config[name] = default

        # push it into the variables
        variables[name] = variable_template(fab.env.config[name], variables)

    default('source_dir', '{deploy_dir}/source')
    default('releases_dir', '{deploy_dir}/releases')
    default('current_symlink', '{deploy_dir}/current')
    default('keep_releases', 10)

    # process variables
    if 'variables' in fab.env.config:
        for definition in fab.env.config['variables']:
            for name, val in definition.iteritems():
                variables[name] = variable_template(val, variables)

    fab.env.config['variables'] = variables


@fab.task
def show(target, config_file):
    '''
    Loads and pretty prints the specified targets config (for debugging)
    '''
    load(target, config_file)

    import pprint
    pprint.pprint(fab.env.config)



import fabric.api as fab

from .config import variable_template


def run(commands, prefix=None, cd=None, shell_env=None):
    """
    This is a shell around the fabric run command that allows for conditional
    use of the prefix, cd & shell_env context managers.

    :param commands: A list of commands to run
    :param prefix: An optional prefix to use
    :param cd:  An optional working directory
    :param shell_env: An optional dict of shell env variables
    :return:
    """
    def _run():
        "closure to make running the commands reusable"
        for command in commands:
            fab.run(command)

    # this is so ugly, but I couldn't come up with a better way to do it
    # without making things terribly complicated
    # TODO make this better somehow
    # XXX maybe this can help: http://stackoverflow.com/a/5359988
    if prefix is not None and cd is not None and shell_env is not None:
        with fab.cd(cd):
            with fab.prefix(prefix):
                with fab.shell_env(**shell_env):
                    _run()
    elif prefix is not None and shell_env is not None:
        with fab.prefix(prefix):
            with fab.shell_env(**shell_env):
                _run()
    elif cd is not None and shell_env is not None:
        with fab.cd(cd):
            with fab.shell_env(**shell_env):
                _run()
    elif cd is not None and prefix is not None:
        with fab.cd(cd):
            with fab.prefix(prefix):
                _run()
    elif cd is not None:
        with fab.cd(cd):
            _run()
    elif prefix is not None:
        with fab.prefix(prefix):
            _run()
    elif shell_env is not None:
        with fab.shell_env(**shell_env):
            _run()
    else:
        _run()


def execute(category, release):
    """
    Run all stages in the specified category for the specified release

    :param category: The category of stages to run (before or after)
    :param release: The name of the release on the host
    :return: None
    """
    if 'stages' not in fab.env.config:
        return fab.warn("No stages defined in your config")

    if category not in fab.env.config['stages']:
        return fab.warn("No stages defined in the '%s' category" % category)

    for stage in fab.env.config['stages'][category]:
        if 'id' not in stage:
            fab.warn("No 'id' defined for stage: %s" % stage)
            continue

        if 'commands' not in stage:
            fab.warn("No 'commands' defined for stage: %s" % stage)
            continue

        if not isinstance(stage['commands'], (list, tuple)):
            fab.warn("The supplied commands are no in the correct format: %s" % stage['commands'])
            continue

        roles = stage.get('roles')

        fab.puts("\nRunning stage: %s (roles: %s)" % (stage['id'],
                                                      ", ".join(roles)))

        prefix = variable_template(stage.get('prefix'))
        cd = variable_template(stage.get('cd'))

        if 'shell_env' in stage:
            shell_env = {}
            for env_name in stage['shell_env']:
                shell_env[env_name] = variable_template(stage['shell_env'][env_name])
        else:
            shell_env = None

        commands = [
            variable_template(command)
            for command in stage['commands']
        ]

        fab.execute(run, commands, prefix, cd, shell_env,
                    roles=roles)



import fabric.api as fab

import iso8601

from . import config
from . import release
from . import stages

__all__ = ('deploy', 'list', 'rollback')


@fab.task
def deploy(target, config_file="deploy.conf"):
    """
    Deploy your application to the specified 'target' in deploy.conf
    """
    # load our config file
    config.load(target, config_file)

    # get our release name
    release_name = release.name(target)

    fab.puts("Deploying new release: %s" % release_name)

    # create the release on the hosts
    fab.execute(release.create, release_name, role='all')

    # run our before tasks for this release
    stages.execute("before", release_name)

    # update the current symlink
    fab.execute(release.make_current, release_name, role='all')

    # run our after tasks for this release
    stages.execute("after", release_name)

    # clean up the releases directory
    fab.execute(release.cleanup, role='all')


def get_releases():
    with fab.hide('output', 'running'):
        ret = fab.execute(release.list, role='all')

    # calculate the intersection of the results
    ret_lists = [
        r[1]
        for r in ret.items()
    ]
    releases = set.intersection(*[
        set(r)
        for r in ret_lists
    ])

    # sort the releases
    releases = [r for r in releases]
    releases.sort()

    return releases


@fab.task
def list(target, config_file="deploy.conf"):
    """
    Lists the releases available on your hosts
    """
    # load our config file
    config.load(target, config_file)

    releases = get_releases()

    fab.puts("Available releases:")
    for rel in releases:
        try:
            ts, ref = rel.split("-")
            iso8601_date = iso8601.parse_date(ts)
            release_name = " ".join([
                "-",
                iso8601_date.strftime("%a, %d %b %Y %H:%M:%S +0000"),
                "GIT reference",
                ref
            ])
        except ValueError:
            release_name = "- No info available"

        fab.puts("* {rel} {release_name}".format(
            rel=rel,
            release_name=release_name
        ))


@fab.task
def rollback(target, release, config_file="deploy.conf"):
    # load our config file
    config.load(target, config_file)

    releases = get_releases()

    if release not in releases:
        fab.abort("The release {release} is not valid".format(
            release=release
        ))

    fab.puts("Rolling back to release: {release}".format(
        release=release
    ))

    # run our rollback tasks for this release
    stages.execute("rollback", release)

    # update the current symlink
    fab.execute(release.make_current, release, role='all')

    # run our after tasks for this release
    stages.execute("after", release)



#!/usr/bin/env python

from setuptools import setup

import re

# load our version from our init file
init_data = open('garment/__init__.py').read()
matches = re.search(r"__version__ = '([^']+)'", init_data, re.M)
if matches:
    version = matches.group(1)
else:
    raise RuntimeError("Unable to load version")

setup(name='garment',
      packages=['garment'],
      scripts=['scripts/deploy'],
      include_package_data=True,
      version=version,
      license="Apache License, Version 2.0",
      description='A collection of fabric tasks that roll up into a single deploy function. The whole process is coordinated through a single deployment configuration file named deploy.conf',
      long_description=open('README.rst').read(),
      author='Evan Borgstrom',
      author_email='evan@borgstrom.ca',
      url='https://github.com/borgstrom/garment',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      install_requires=[
          'setuptools',
          'PyYAML',
          'fabric>=1.8,<=1.9',
          'iso8601>=0.1,<=0.2'
      ])
